'''
Created on 23.12.2015

@author: johanneskinzig
@author: christianvonharscher

@version: 0.1
@versionDescription: learns faces by name and recognizes them by name and a confidence value
'''

""" This sample is only working with OpenCV Version 2.4...
OpenCV Version 3 does not support the FaceRecognizer. If you still want to use OPpenCV 3 you have to compile opencv_contrib additionally.
"""

import cv2
import numpy

from collections import Counter

# benchmarking and analysis
from timemeasurement import *


class DetectionRecognition:
    """ Is required for the face recognition and detection."""
    def __init__(self, cascPath, videodevice, height, width):
        # videodevices which are connected to the computer
        self.videodevice = videodevice
        # set resolution
        self.cameraResHeight = height
        self.cameraResWidth = width
        # path to face classifier
        self.faceCascade = cv2.CascadeClassifier(cascPath)
        # For face recognition the LBPHF recognizer will be used
        self.recognizer = cv2.createLBPHFaceRecognizer()
        
        ######### For FaceTrainer: ###########
        # arrays for learned faces and labels
        # stores learned faces
        # both will be handed over to the trainer --> Data contained there has to be DEFENETELY IN SYNC!!!
        self.feature_faces = []
        # stores label for faces
        self.feature_faces_labels = []
        
        ######### For Mapping Table: ###########
        # the trainer expects the labels as integers but the user wants to see the face's name
        # this means a mapping table is necessary
        self.feature_name_label_mapping = []
        
        
        ######### For the time measurement (benchmarking): ###########
        # instanciating the benchmarker for the FaceDetector
        self.faceDetectorBenchmarker = TimeMeter("Detector")
        self.faceDetectorBenchmarker.initHeader("wallclock time;\t cpu time;\t camera resolution\n")

        # instanciating the benchmarker for the FaceRecognizer
        self.faceRecognizerBenchmarker = TimeMeter("Recognizer")
        self.faceRecognizerBenchmarker.initHeader("DistinctNumberOfFacesInTrainingSet;\t SamplesPerFace;\t TotalNumberOfFacesInTrainingSet;\t wallclock time;\t cpu time;\t camera resolution\n")


        
    def getCameraResolution(self):
        """Returns the set camera resolution as string."""
        return str(self.cameraResHeight) + "x" + str(self.cameraResWidth)
        
    
    def detectFaces(self):
        """ Detects the face and draws a rectangle around it. Depending on the user's input the face is stored for recognition."""
        # access camera
        self.videoStream = cv2.VideoCapture(self.videodevice)
        # resize stream
        self.videoStream.set(3,self.cameraResHeight)
        self.videoStream.set(4,self.cameraResWidth)
        
        while True:
            # Capture frame-by-frame
            self.ret, self.frame = self.videoStream.read()
            
            # convert image to gray, because recognition is done in gray
            self.grayframe = cv2.cvtColor(self.frame, cv2.COLOR_BGR2GRAY)
            
            self.faceDetectorBenchmarker.startBenchmark() # --> starts the benchmark
            # contains an array of rectangular coordinates that surround each face which is found in the image
            self.faces = self.faceCascade.detectMultiScale(self.grayframe, scaleFactor=1.2, minNeighbors=5, minSize=(30, 30), flags=cv2.cv.CV_HAAR_SCALE_IMAGE)
            
            self.walltime, self.cputime = self.faceDetectorBenchmarker.stopBenchmark() # --> stops the benchmark
            self.faceDetectorBenchmarker.writeToFile(str(self.walltime) + ";\t" + str(self.cputime) + ";\t" + self.getCameraResolution() + "\n") # --> writes the gathered data to file
            
            # draw rectangle around face
            for (x, y, w, h) in self.faces:
                cv2.rectangle(self.frame, (x, y), (x+w, y+h), (0, 255, 0), 2)
                
            
            cv2.putText(self.frame,"FaceTrainer", (6,20), cv2.FONT_HERSHEY_TRIPLEX, 0.75, (0, 255, 0))
            cv2.putText(self.frame,"a --> add Face to training set ", (6,45), cv2.FONT_HERSHEY_TRIPLEX, 0.75, (0, 255, 0))
            cv2.putText(self.frame,"r --> switch to FaceRecognizer", (6,70), cv2.FONT_HERSHEY_TRIPLEX, 0.75, (0, 255, 0))
            
            # Display the videoStream and the features
            cv2.imshow('FaceTrainer', self.frame)
            
            # Enter name for face when receiving 'a' --> add
            if (cv2.waitKey(50) & 255) == 97:
                self.addFaceToTrainingSet(self.grayframe, x, y, w, h)
            
            # Quit window and call training method when received 'r'
            if (cv2.waitKey(50) & 255) == 114:
                print "-- Switching - Training recognizer now --"
                # call trainer
                self.trainRecognizerWithFaces(self.feature_faces, self.feature_faces_labels)
                print "-- Done --"
                # quit current windows
                cv2.destroyAllWindows()
                break
            
    
    def addFaceToTrainingSet(self, image, x, y, w, h):
        """ Ads faces to the training database which is later handed over to the recognizer.train method."""
        self.feature_face_name = raw_input("Spell name for current face: ")
        
        if self.feature_face_name in self.feature_name_label_mapping:
            self.feature_faces.append(image[y: y + h, x: x + w])
            self.feature_faces_labels.append(self.feature_name_label_mapping.index(self.feature_face_name))
            print ("Added object",  self.feature_name_label_mapping.index(self.feature_face_name), "to training set, with name ", self.feature_face_name) 
        else:
            self.feature_name_label_mapping.append(self.feature_face_name)
            self.feature_faces.append(image[y: y + h, x: x + w])
            self.feature_faces_labels.append(self.feature_name_label_mapping.index(self.feature_face_name))
            print ("Added object",  self.feature_name_label_mapping.index(self.feature_face_name), "to training set, with name ", self.feature_face_name)
            
    
    def showCurrentFacesInTrainingSetAsEnumeration(self):
        """Returns the content of the labels array. Enumerates the content."""
        print "Faces already added: ", self.feature_faces_labels
        return self.feature_faces_labels
        
    def showCurrentFacesInTrainingSetAsTotal(self):
        """Returns the length of the labels array. Returns the total number of faces in training set."""
        print "Number of faces in training set: ", len(self.feature_faces_labels)
        return len(self.feature_faces_labels)
        
    def showCurrentFacesInTrainingSetAsDistinctNumberOfFaces(self):
        """Returns the distinct faces in training set."""
        self.distinctNumberOfFaces = set(self.feature_faces_labels)
        self.distinctNumberOfFaces = list(self.distinctNumberOfFaces)
        print "Distinct number of faces in training set: ", len(self.distinctNumberOfFaces)
        return len(self.distinctNumberOfFaces)
        
    def showNumberOfSamplesPerFace(self):
        """Returns the number of samples per face which are currently contained in training set.""" 
        self.collection = self.feature_faces_labels
        self.samplesPerFaceDict = Counter(self.collection)
        print "Features per face: ",  self.samplesPerFaceDict
        return self.samplesPerFaceDict
        
    def trainRecognizerWithFaces(self, features, labels):
        """ Takes the OpenCV FaceRecognizer and calls the "train" method.
            This takes the training set and hands it over to the recognizing trainer.
        """
        self.recognizer.train(features, numpy.array(labels))
        
    def recognizeFaces(self):
        """ Recognize faces and display their labels, names."""
        while True:
            # Capture frame-by-frame
            self.ret, self.frame = self.videoStream.read()
            # convert to gray again to detect faces
            self.grayframe = cv2.cvtColor(self.frame, cv2.COLOR_BGR2GRAY)
            # detect faces
            self.faces = self.faceCascade.detectMultiScale(self.grayframe, scaleFactor=1.2, minNeighbors=5, minSize=(30, 30), flags=cv2.cv.CV_HAAR_SCALE_IMAGE)
    
            # Draw a rectangle around the faces
            for (x, y, w, h) in self.faces:
                # draw rectangle around faces
                cv2.rectangle(self.frame, (x, y), (x+w, y+h), (0, 255, 0), 2)
                
                self.faceRecognizerBenchmarker.startBenchmark() # --> starts the benchmark
                # perform the face recognition
                self.predicted, self.conf = self.recognizer.predict(self.grayframe[y: y + h, x: x + w])
                
                self.walltime, self.cputime = self.faceRecognizerBenchmarker.stopBenchmark() # --> stops the benchmark
                self.faceRecognizerBenchmarker.writeToFile(str(self.showCurrentFacesInTrainingSetAsDistinctNumberOfFaces()) + ";\t" + str(self.showNumberOfSamplesPerFace()) + ";\t" + str(self.showCurrentFacesInTrainingSetAsTotal()) + ";\t" + str(self.walltime) + ";\t" + str(self.cputime) + ";\t" + self.getCameraResolution() + "\n") # --> writes the gathered data to file
                
                # put name to face
                cv2.putText(self.frame, self.feature_name_label_mapping[self.predicted], (x, y-10), cv2.FONT_HERSHEY_TRIPLEX, 0.75, (0, 255, 0))
                
                # print the debug information to stdout
                print "predicted: ", self.predicted, "confidence: ", self.conf
                print "Face: ", self.feature_name_label_mapping[self.predicted], "confidence: ", self.conf
                
            
            cv2.putText(self.frame,"FaceRecognizer", (6,20), cv2.FONT_HERSHEY_TRIPLEX, 0.75, (0, 255, 0))
            cv2.putText(self.frame,"q --> Quit Application ", (6,45), cv2.FONT_HERSHEY_TRIPLEX, 0.75, (0, 255, 0))
            # Display the videoStream and the features
            cv2.imshow('FaceRecognizer', self.frame)

            # check for key press 'q' --> quit
            if (cv2.waitKey(50) & 255) == 113:
                # When everything is done, release the capture
                self.videoStream.release()
                cv2.destroyAllWindows()
                break

        
if __name__ == "__main__":
    cascPath = "haarcascade_frontalface_default.xml"
    computerVision = DetectionRecognition(cascPath, 0, 800, 600)

    # detect and learn faces first
    computerVision.detectFaces()
    print(computerVision.showCurrentFacesInTrainingSetAsDistinctNumberOfFaces())
    print(computerVision.showCurrentFacesInTrainingSetAsTotal())
    print(computerVision.showNumberOfSamplesPerFace())

    # when training has finished recognize faces
    computerVision.recognizeFaces()

    # quit
    print "Application was shut down"
