'''
Created on 05.01.2016

@author: johanneskinzig
@version: 0.1

@versionDescription: measures wall clock time and CPU execution time 
'''

import time

class TimeMeter():
    """"Used for analyzing and benchmarking the timing constraints of code snippets.
    The runtime is measured in seconds. The gathered data is written to a CSV file."""

    def __init__(self, name):
        """Measures the CPU execution and wall clock time (CPU execution time only works properly on Unix like systems)."""
        # set name for instance
        self.name = name
        # unique filename for every instance based on date, time and instance name
        self.outFilepath = "benchmarking_data/" + time.strftime("%Y-%m-%dT-%H-%M-%S", time.gmtime()) + "_" + self.name + ".csv"

    def initHeader(self, HeaderAsCSV):
        """Writes the headers to the file. Columns separated by comma."""
        # If the file does not exist, it creates a new file for writing
        self.outFile = open(self.outFilepath, "a")
        self.outFile.write(HeaderAsCSV)
        self.outFile.close()
        
    def writeToFile(self, StringAsCSV):
        """Write data to file. Columns separated by comma."""
        self.outFile = open(self.outFilepath, "a")
        self.outFile.write(StringAsCSV)
        self.outFile.close

    def startCPUMeter(self):
        """Starts the CPU execution measurement"""
        self.start_point_in_time_cpu = time.clock()
    
    def stopCPUMeter(self):
        """"Stops the CPU execution measurement and returns the measured values"""
        self.time_consumed_cpu = time.clock() - self.start_point_in_time_cpu
        #print "CPU time measured by ", self.name, ": ", self.time_consumed_cpu, " seconds"
        return self.time_consumed_cpu
    
    def startWallClockMeter(self):
        """Starts the wall clock measurement."""
        self.start_point_in_time_wall = time.time()
        
    def stopWallClockMeter(self):
        """Stops the wall clock measurement and returns the measured values."""
        self.time_consumed_wall = time.time() - self.start_point_in_time_wall
        #print "Wall clock time measured by ", self.name, ": ", self.time_consumed_wall, " seconds"
        return self.time_consumed_wall
    
    def startBenchmark(self):
        """Starts the wallclock and the cpu measurement clock."""
        self.startWallClockMeter()
        self.startCPUMeter()

    def stopBenchmark(self):
        """Stops the benchmarking (the two running clocks) and returns the data directly."""
        walltime = self.stopWallClockMeter()
        cputime = self.stopCPUMeter()
        return walltime, cputime
    
if __name__ == "__main__":
    # sample
    measurement1 = TimeMeter("Measurement1")
    
    # measure cpu execution time
    measurement1.startBenchmark()
    time.sleep(1)
    wallclock, cpuclock = measurement1.stopBenchmark()
    measurement1.writeToFile(str(wallclock) + ",\t" + str(cpuclock) + "\n")
    
    measurement2 = TimeMeter("Measurement2")
    measurement2.initHeader("faces in training set,\t distinct faces in training set,\t number of samples per face,\t wallclock,\t cpu time\n")
    measurement2.startBenchmark()
    time.sleep(1)
    wall1, cpuclock1 = measurement2.stopBenchmark()
    measurement2.writeToFile(str(wall1) + ",\t" + str(cpuclock1) + "\n")
    